# -*- coding: utf-8 -*-

from __future__ import absolute_import
import threading
from datetime import datetime, timedelta
from .default_consumer import DefaultConsumer


class FlushThread(threading.Thread):
    '''
    FlushThread is used to asynchronously flush the events stored in
    the AsyncBufferedConsumer buffers.
    '''

    def __init__(self, consumer):
        threading.Thread.__init__(self)
        self.consumer = consumer

    def run(self):
        self.consumer._sync_flush()


class AsyncBufferedConsumer(DefaultConsumer):

    def __init__(self, product_id, data_source_id, server_host,
                 flush_after=timedelta(0, 10), flush_first=False,
                 max_size=50, request_timeout=2,
                 retry_limit=3, retry_backoff_factor=0.25, verify_cert=True):

        super(AsyncBufferedConsumer, self).__init__(
            product_id, data_source_id, server_host,
            request_timeout=request_timeout,
            retry_limit=retry_limit,
            retry_backoff_factor=retry_backoff_factor,
            verify_cert=verify_cert,
        )

        # remove the minimum max size that the SynchronousBufferedConsumer
        self._max_size = max_size
        self.flush_after = flush_after
        self.flush_first = flush_first

        self._async_events = []
        self._async_items = []

        self._async_buffers = {
            'events': [],
            'items': [],
        }

        if not self.flush_first:
            self.last_flushed = datetime.now()
        else:
            self.last_flushed = None

        self.flush_lock = threading.Lock()
        self.flushing_thread = None

    def _flush_thread_is_free(self):
        '''
        Check whether a thread is currently being used to flush events. This
        guarantees that only one thread is ever used at a time to flush.
        '''
        return self.flushing_thread is None or not self.flushing_thread.is_alive()

    def _should_flush(self):
        '''
        Checks whether the events in the AsyncBufferedConsumer should be flushed.
        '''
        full = len(self._async_events) >= self._max_size or len(self._async_items) > 0

        # always flush the first event
        stale = self.last_flushed is None

        if not stale and self.flush_after:
            stale = datetime.now() - self.last_flushed > self.flush_after

        if stale:
            return True

        if full:
            return True

        return False

    def send(self, message):

        if hasattr(message, 'event_type'):
            buf = self._async_events
            buf.append(message)
        else:
            buf = self._async_items
            buf.append(message)

        should_flush = self._should_flush()
        if should_flush:
            self.flush()

    def flush(self):

        with self.flush_lock:
            if self._flush_thread_is_free():

                self.transfer_buffers()

                self.flushing_thread = FlushThread(self)
                self.flushing_thread.start()

                flushing = True
            else:
                flushing = False

        if flushing:
            self.last_flushed = datetime.now()

        return flushing

    def transfer_buffers(self):
        buf = self._async_events
        while buf:
            self._async_buffers['events'].append(buf.pop(0))
        buf = self._async_items
        while buf:
            self._async_buffers['items'].append(buf.pop(0))

    def _sync_flush(self):
        if len(self._async_buffers['events']) > 0:
            request_url = self.endpoints['collect']
            buf = self._async_buffers['events']
            while buf:
                batch = buf[:self._max_size]
                data = self._data_parse.get_bytes(batch)
                self.send_data(
                    request_url,
                    data,
                    headers={'content-type': 'application/json'}
                )
                buf = buf[self._max_size:]
            self._async_buffers['events'] = buf

        if len(self._async_buffers['items']) > 0:
            request_url = self.endpoints['item']
            buf = self._async_buffers['items']
            while buf:
                batch = buf[:self._max_size]
                data = self._data_parse.get_bytes(batch)
                self.send_data(
                    request_url,
                    data,
                    headers={'content-type': 'application/json'}
                )
                buf = buf[self._max_size:]
            self._async_buffers['items'] = buf



# -*- coding: utf-8 -*-
from .default_consumer import DefaultConsumer


class BufferedConsumer(DefaultConsumer):

    def __init__(self, product_id, data_source_id, server_host, max_size=50,
                 retry_limit=3, request_timeout=2, retry_backoff_factor=0.25, verify_cert=True):
        super(BufferedConsumer, self).__init__(product_id, data_source_id, server_host,
                                               retry_limit, request_timeout, retry_backoff_factor, verify_cert)
        self._events = []
        self._max_size = min(200, max_size)

    def send(self, message):
        if hasattr(message, 'event_type'):
            request_url = self.endpoints['collect']
            buf = self._events
            buf.append(message)
            if len(buf) >= self._max_size:
                self._flush_event(request_url)
        else:
            request_url = self.endpoints['item']
            messages = [message]
            data = self._data_parse.get_bytes(messages)
            self.send_data(
                request_url,
                data,
                headers={'content-type': 'application/json'}
            )

    def _flush_event(self, request_url):
        buf = self._events

        while buf:
            batch = buf[:self._max_size]
            data = self._data_parse.get_bytes(batch)
            self.send_data(
                request_url,
                data,
                headers={'content-type': 'application/json'}
            )
            buf = buf[self._max_size:]
        self._events = buf

    def flush(self):
        request_url = self.endpoints['collect']
        self._flush_event(request_url)



# -*- coding: utf-8 -*-

from .data_parser import JsonParser
from .consumer import Consumer


class DefaultConsumer(Consumer):
    def __init__(self, product_id, data_source_id, server_host,
                 retry_limit=3, request_timeout=2, retry_backoff_factor=0.25, verify_cert=True):
        super(DefaultConsumer, self).__init__(product_id, data_source_id, server_host,
                                              retry_limit, request_timeout, retry_backoff_factor, verify_cert)
        self._data_parse = JsonParser(product_id, data_source_id)

    def send(self, message):
        if hasattr(message, 'event_type'):
            request_url = self.endpoints['collect']
        else:
            request_url = self.endpoints['item']

        messages = [message]
        data = self._data_parse.get_bytes(messages)

        self.send_data(
            request_url,
            data,
            headers={'content-type': 'application/json'}
        )



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals
import json
from .version import VERSION


class DataParser(object):
    def get_bytes(self, messages):
        pass


class JsonParser(DataParser):

    def __init__(self, product_id, data_source_id):
        self._product_id = product_id
        self._data_source_id = data_source_id

    def get_bytes(self, messages):
        batch = []
        for message in messages:
            batch.append(self.json(message))
        batch_json = '[{0}]'.format(','.join(batch))
        return batch_json

    @staticmethod
    def json_dumps(data, cls=None):
        return json.dumps(data, separators=(',', ':'), cls=cls)

    def json(self, message):
        dict = {}

        # event message
        if hasattr(message, 'event_type') and message.event_type is not None:
            dict['eventType'] = message.event_type
        if hasattr(message, 'event_time') and message.event_time is not None:
            dict['timestamp'] = message.event_time
        if hasattr(message, 'event_name') and message.event_name is not None:
            dict['eventName'] = message.event_name
        if hasattr(message, 'anonymous_id') and message.anonymous_id is not None:
            dict['deviceId'] = message.anonymous_id
        if hasattr(message, 'login_user_key') and message.login_user_key is not None:
            dict['userKey'] = message.login_user_key
        if hasattr(message, 'login_user_id') and message.login_user_id is not None:
            dict['userId'] = message.login_user_id
        if hasattr(message, 'attributes') and message.attributes is not None:
            dict['attributes'] = message.attributes

        # item message
        if hasattr(message, 'key') and message.key is not None:
            dict['key'] = message.key
        if hasattr(message, 'id') and message.id is not None:
            dict['id'] = message.id
        if hasattr(message, 'attrs') and message.attrs is not None:
            dict['attributes'] = message.attrs

        # common
        dict['dataSourceId'] = self._data_source_id
        dict['projectKey'] = self._product_id

        dict['sdkVersion'] = VERSION

        return self.json_dumps(dict)
